#include "stdafx.h"
#include "nstest.h"
#include "ddeobj.h"

#include "nstesdoc.h"
#include "stdarg.h"

//	Serialization, run time info, dyn create.
IMPLEMENT_SERIAL( CDDEObject, CObject, 1 )

//	Initialize static members.
CObList CDDEObject::m_colRunning;
int CDDEObject::m_iInstanceCounter = 0;
DWORD CDDEObject::m_dwidInst;
HSZ CDDEObject::m_aTopics[CDDEObject::m_MaxTopics];
HSZ CDDEObject::m_hszBrowser;

//	Construction
CDDEObject::CDDEObject()	{
	m_pDoc = NULL;
	CommonConstruction();
}

CDDEObject::CDDEObject(CNstestDoc *pDoc)	{
	m_pDoc = pDoc;
	CommonConstruction();
}

void CDDEObject::CommonConstruction()	{
	//	Init everything but the doc.
	//	That's been done in the actual constructor.
	
	//	Next instance please.
	m_iInstanceCounter++;
	TRACE("Instance count is %d.\n", m_iInstanceCounter);
	
	//	Create the DDE service name that we'll be using in this object.
	char aBuf[20];
	_itoa(m_iInstanceCounter, aBuf, 10);
	m_csServiceName.LoadString(IDS_DDE_SERVICE_NAME);
	m_csServiceName += aBuf;
	TRACE("Service name is %s\n", (const char *)m_csServiceName);
	
	//	Add us to the list of running objects.
	m_rIndex = m_colRunning.AddTail(this);
		
	//	If this is the only CDDEObject, then we need to load all topics
	m_hszServiceName = NULL;
	if(m_colRunning.GetCount() == 1)	{
		LoadTopics();
		
		//	Get the service name of the browser we're communicating with.
		CString csBrowser;
		csBrowser.LoadString(IDS_BROWSER_SERVICE);
		m_hszBrowser = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csBrowser, CP_WINANSI);
		GetDocument()->Display(ExplainError());
	}
	
	//	Register our service name.
	m_hszServiceName = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)m_csServiceName,
		CP_WINANSI);
	GetDocument()->Display(ExplainError());		
	DdeNameService(m_dwidInst, m_hszServiceName, NULL, DNS_REGISTER);
	GetDocument()->Display(ExplainError());
}

//	Destruction
CDDEObject::~CDDEObject()	{
	//	Disconnect every conversation that this object is handling.
	POSITION rTraverse = m_colConversations.GetHeadPosition();
	CDDEConversation *pConv;
	while(rTraverse != NULL)	{
		pConv = (CDDEConversation *)m_colConversations.GetNext(rTraverse);
		pConv->DoDisconnect();
	}

	//	Take ourselves out of the running object list.
	m_colRunning.RemoveAt(m_rIndex);
	
	//	Unregister our service name.
	DdeNameService(m_dwidInst, m_hszServiceName, NULL, DNS_UNREGISTER);
	GetDocument()->Display(ExplainError());
	DdeFreeStringHandle(m_dwidInst, m_hszServiceName);
	GetDocument()->Display(ExplainError());
	m_hszServiceName = NULL;
	
	//	If there are no more running objects, it's time to flush all the topics.
	if(m_colRunning.GetCount() == 0)	{	
		FlushTopics();
		
		DdeFreeStringHandle(m_dwidInst, m_hszBrowser);
		GetDocument()->Display(ExplainError());
	}
}

//	Lookup
CDDEObject *CDDEObject::ResolveService(const char *pServiceName)	{
	//	We need to go through each object that we have, and see if it is
	//		the service that we are looking for.
	POSITION rTraverse = m_colRunning.GetHeadPosition();
	CDDEObject *pTraverse = NULL;
	while(rTraverse != NULL)	{
		pTraverse = (CDDEObject *)m_colRunning.GetNext(rTraverse);
		if(pTraverse->m_csServiceName == pServiceName)	{
			break;
		}
		else	{
			pTraverse = NULL;
		}
	}
	
	return(pTraverse);
}
CDDEObject *CDDEObject::ResolveService(HSZ hszServiceName)	{
	//	Convert the HSZ to a string to perform the lookup.
	DWORD dwLength = DdeQueryString(m_dwidInst, hszServiceName, NULL, 0,
		CP_WINANSI) + 1;
	char *pTemp = new char[dwLength];
	DdeQueryString(m_dwidInst, hszServiceName, pTemp, dwLength, CP_WINANSI);
	
	CDDEObject *pRetVal = ResolveService(pTemp);
	delete pTemp;
	return(pRetVal);
}
CDDEObject *CDDEObject::ResolveConversation(CDDEConversation *pConv)	{
	//	We need to go through each object that we have, and see if it owns
	//		the conversation that we are looking for.
	POSITION rTraverse = m_colRunning.GetHeadPosition();
	CDDEObject *pTraverse = NULL;
	POSITION rTravConv = NULL;
	CDDEConversation *pTravConv = NULL;
	while(rTraverse != NULL)	{
		pTraverse = (CDDEObject *)m_colRunning.GetNext(rTraverse);
		
		rTravConv = pTraverse->m_colConversations.GetHeadPosition();
		while(rTravConv != NULL)	{
			pTravConv =
				(CDDEConversation *)pTraverse->m_colConversations.GetNext(rTravConv);
			if(pTravConv == pConv)	{
				return(pTraverse);
			}
		}		
	}
	
	return(NULL);
}

//	Connection management.
HDDEDATA CDDEObject::ServerConnect(HSZ hszTopic)	{
	//	just accept anything....
	return(DDETRUE);
}

//	Connection confirmation.
void CDDEObject::ServerConnectConfirm(HCONV hConv)	{
	//	Create a new conversation object, and add it to the conversations
	//		that this oject is handling.
	CDDEConversation *pConversation = new CDDEConversation(hConv);
	m_colConversations.AddTail(pConversation);
}

//	Disconnection of a conversation managment.
void CDDEObject::DDEDisconnect(CDDEConversation *pConv)	{
	//	Go through every running DDE object, go through ever conversation,
	//		until we find the correct conversation, and then remove it from
	//		the conversation list.
	//	Actual disconnection is taking place elsewhere....
	POSITION rTravObj = m_colRunning.GetHeadPosition();
	CDDEObject *pObject;
	POSITION rTravConv, rRemoveMe;
	while(rTravObj != NULL)	{
		pObject = (CDDEObject *)m_colRunning.GetNext(rTravObj);
		rTravConv = pObject->m_colConversations.GetHeadPosition();
		while(rTravConv != NULL)	{
			rRemoveMe = rTravConv;
			if(pConv == (CDDEConversation *)pObject->m_colConversations.GetNext(rTravConv))	{
				pObject->m_colConversations.RemoveAt(rRemoveMe);
				return;
			}
		}
	}
}

//	WWW topics
void CDDEObject::LoadTopics()	{
	CString csTopic;
	
	csTopic.LoadString(IDS_WWW_ACTIVATE);
	m_aTopics[m_Activate] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_ALERT);
	m_aTopics[m_Alert] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_BEGINPROGRESS);
	m_aTopics[m_BeginProgress] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_CANCELPROGRESS);
	m_aTopics[m_CancelProgress] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_ENDPROGRESS);
	m_aTopics[m_EndProgress] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_EXIT);
	m_aTopics[m_Exit] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_GETWINDOWINFO);
	m_aTopics[m_GetWindowInfo] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_LISTWINDOWS);
	m_aTopics[m_ListWindows] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_MAKINGPROGRESS);
	m_aTopics[m_MakingProgress] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_OPENURL);
	m_aTopics[m_OpenURL] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_PARSEANCHOR);
	m_aTopics[m_ParseAnchor] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_QUERYURLFILE);
	m_aTopics[m_QueryURLFile] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_QUERYVIEWER);
	m_aTopics[m_QueryViewer] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_REGISTERPROTOCOL);
	m_aTopics[m_RegisterProtocol] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_REGISTERURLECHO);
	m_aTopics[m_RegisterURLEcho] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_REGISTERVIEWER);
	m_aTopics[m_RegisterViewer] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_REGISTERWINDOWCHANGE);
	m_aTopics[m_RegisterWindowChange] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_SETPROGRESSRANGE);
	m_aTopics[m_SetProgressRange] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_SHOWFILE);
	m_aTopics[m_ShowFile] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_UNREGISTERPROTOCOL);
	m_aTopics[m_UnRegisterProtocol] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_UNREGISTERURLECHO);
	m_aTopics[m_UnRegisterURLEcho] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_UNREGISTERVIEWER);
	m_aTopics[m_UnRegisterViewer] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_UNREGISTERWINDOWCHANGE);
	m_aTopics[m_UnRegisterWindowChange] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_URLECHO);
	m_aTopics[m_URLEcho] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_VERSION);
	m_aTopics[m_Version] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_VIEWDOCFILE);
	m_aTopics[m_ViewDocFile] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	csTopic.LoadString(IDS_WWW_WINDOWCHANGE);
	m_aTopics[m_WindowChange] = DdeCreateStringHandle(m_dwidInst, (char *)(const char *)csTopic, CP_WINANSI);
	GetDocument()->Display(ExplainError());
}

void CDDEObject::FlushTopics()	{
	for(int iTraverse = 0; iTraverse < m_MaxTopics; iTraverse++)	{
		DdeFreeStringHandle(m_dwidInst, m_aTopics[iTraverse]);
		GetDocument()->Display(ExplainError());
	}
}

int CDDEObject::EnumTopic(HSZ hszTopic)	{
	//	Simply go through our held topics, and return the item which matches.
	//	Since HSZs are held by the system, handles should always match up.
	for(int iCounter = 0; iCounter < m_MaxTopics; iCounter++)	{
		if(hszTopic == m_aTopics[iCounter])	{
			break;
		}
	}
	
	if(iCounter != m_MaxTopics)	{
		return(iCounter);
	}
	return(-1);
}

//	Useful client/server methods.
CDDEConversation *CDDEObject::ClientConnect(int iTopic)	{
	ASSERT(iTopic < m_MaxTopics);	
	HSZ hszTopic = m_aTopics[iTopic];
	
	TRACE("Attempting conversation %d\n", iTopic);
	
	HCONV hConv = DdeConnect(m_dwidInst, m_hszBrowser, hszTopic, NULL);
	GetDocument()->Display(ExplainError());
	
	CDDEConversation *pRetVal = NULL;
	if(hConv != NULL)	{
		pRetVal = new CDDEConversation(hConv);
		m_colConversations.AddTail(pRetVal);
	}
	
	return(pRetVal);
}

HSZ CDDEObject::ClientArguments(const char *pFormat, ...)	{
	//	Always pass in pointer values.
	//	This way, a NULL, can be interpreted as an optional, missing parameter.
	char *pTraverse = (char *)pFormat;	

	//	Variable number of arguments.
	va_list VarList;
	va_start(VarList, pFormat);

	char caNumpad[64];
	CString csBuffer;
	CString csRetval;
	
	while(*pTraverse)	{
		//	Erase temp data from our last pass.
		caNumpad[0] = '\0';
		csBuffer.Empty();
	
		//	Compare our current format to the known formats
		if(0 == strncmp(pTraverse, "DW", 2))	{
			//	A DWORD.
			DWORD *pWord = va_arg(VarList, DWORD *);
			
			if(pWord != NULL)	{
			    //  See if we need to use hex.
			    if(((CNstestApp *)AfxGetApp())->m_bHex) {
				    sprintf(caNumpad, "%#lx", *pWord);
                }
                else    {
				    sprintf(caNumpad, "%lu", *pWord);
                }
				csRetval += caNumpad;
			}
		}
		else if(0 == strncmp(pTraverse, "CS", 2))	{
			//	A CString, not quoted
			CString *pCS = va_arg(VarList, CString *);
			
			if(pCS != NULL)	{
				csRetval += *pCS;
			}
		}
		else if(0 == strncmp(pTraverse, "QCS", 3))	{
			//	A quoted CString
			CString *pQCS = va_arg(VarList, CString *);
			
			if(pQCS != NULL)	{
				csRetval += '\"';
				
				//	Need to escape any '"' to '\"', literally.
				char *pConvert = (char *)(const char *)*pQCS;
				while(*pConvert != '\0')	{
					if(*pConvert == '\"')	{
						csRetval += '\\';
					}
					csRetval += *pConvert;
					pConvert++;
				}
				csRetval += '\"';
			}
		}
		else if(0 == strncmp(pTraverse, "BL", 2))	{
			//	A boolean
			TwoByteBool *pBool = va_arg(VarList, TwoByteBool *);
			
			if(pBool != NULL)	{
				if(*pBool != FALSE)	{
					csRetval += "TRUE";
				}
				else	{
					csRetval += "FALSE";
				}
			}
		}
		else	{
			//	Unhandled format, just get out of loop.
			ASSERT(0);
			break;
		}
		
		//	Go on to next type
		pTraverse = SkipToNextArgument(pTraverse);
		
		//	See if we need a comma
		if(*pTraverse != '\0')	{
			csRetval += ',';
		}
	}

	va_end(VarList);
	
	//	Make sure we're atleast returning something.
	if(csRetval.IsEmpty())	{
		return(NULL);
	}

	//	Return our resultant HSZ, created from our string.
	HSZ Final = DdeCreateStringHandle(m_dwidInst, (char *)
		(const char *)csRetval, CP_WINANSI);
	GetDocument()->Display(ExplainError());
	return(Final);
}

char *SkipToNextArgument(char *pFormat)	{
	//	Safety dance
	if(pFormat == NULL || *pFormat == '\0')	{
		return(pFormat);
	}

	//	The next format is directly after a ','
	while(*pFormat != ',' && *pFormat != '\0')	{
		pFormat++;
	}
	if(*pFormat == ',')	{
		pFormat++;
	}
	
	return(pFormat);
}

char *ExtractArgument(HSZ hszArgs, int iArg)	{
	//	Quoted strings are counted as only one argument, though
	//		the quotes are not copied into the return string.

	DWORD dwLength = DdeQueryString(CDDEObject::m_dwidInst, hszArgs, NULL, 0L,
		CP_WINANSI) + 1;
	char *pTraverse = new char[dwLength];
	char *pRemove = pTraverse;
	if(pTraverse == NULL)	{
		return(NULL);
	}	
	DdeQueryString(CDDEObject::m_dwidInst, hszArgs, pTraverse, dwLength, CP_WINANSI);
	
	//	safety dance
	if(*pTraverse == '\0' || iArg < 1)	{
		delete(pRemove);
		return(NULL);
	}
	
	//	Need to decrement the argument we're looking for, as the very
	//		first argument has no ',' at the beginning.
	iArg--;
	
	//	Search through the arguments, seperated by ','.
	while(iArg)	{
		//	Special handling of quoted strings.
		if(*pTraverse == '\"')	{
			//	Find the ending quote.
			while(*pTraverse != '\0')	{
				pTraverse++;
				if(*pTraverse == '\"')	{
					pTraverse++;	//	One beyond, please
					break;
				}
				else if(*pTraverse == '\\')	{
					//	Attempting to embed a quoted, perhaps....
					if(*(pTraverse + 1) == '\"')	{
						pTraverse++;
					}
				}
			}
		}
		while(*pTraverse != '\0' && *pTraverse != ',')	{
			pTraverse++;
		}

		//	Go beyond a comma
		if(*pTraverse == ',')	{
			pTraverse++;
		}
		
		iArg--;
		
		if(*pTraverse == '\0')	{
			break;
		}
	}
	
	//	Handle empty arguments here.
	if(*pTraverse == ',' || *pTraverse == '\0')	{
		delete(pRemove);
		return(NULL);
	}
	
	int iLength = 1;
	char *pCounter = pTraverse;
	TwoByteBool bQuoted = FALSE;
	
	//	specially handle quoted strings
	if(*pCounter == '\"')	{
		pCounter++;
		bQuoted = TRUE;

		while(*pCounter != '\0')	{
			if(*pCounter == '\"')	{
				break;
			}
			else if(*pCounter == '\\')	{
				if(*(pCounter + 1) == '\"')	{
					pCounter++;
					iLength++;
				}
			}
			
			pCounter++;
			iLength++;
		}
	}
	while(*pCounter != '\0' && *pCounter != ',')	{
		iLength++;
		pCounter++;
	}
	
	//	Subtrace one to ignore ending quote if we were quoted....
	if(bQuoted == TRUE)	{
		iLength--;
	}
	
	//	Argument's of length 1 are of no interest.
	if(iLength == 1)	{
		delete(pRemove);
		return(NULL);
	}
	
	char *pRetVal = new char[iLength];
	
	if(*pTraverse == '\"')	{
		pTraverse++;
	}	
	strncpy(pRetVal, pTraverse, iLength - 1);
	pRetVal[iLength - 1] = '\0';
	
	delete(pRemove);
	return(pRetVal);
}

void CDDEObject::ServerReturned(HDDEDATA hArgs, const char *pFormat, ...)	{
	//	Of course, only pointers should be passed in as the variable number of
	//		arguments so assignment can take place.
	//	hArgs is free'd off by this function.
	char *pDataArgs = (char *)DdeAccessData(hArgs, NULL);
	GetDocument()->Display(ExplainError());
	
	//	Initialize variable number of argumetns.
	va_list VarList;
	va_start(VarList, pFormat);	
	
	//	It will be possible that with only one argument,
	//		that we are receiving raw data intead of string notation.
	if(strchr(pFormat, ',') == NULL)	{
		//	Only one argument.
		if(strcmp(pFormat, "DW") == 0)	{
			DWORD *pWord;
			pWord = va_arg(VarList, DWORD *);
			*pWord = *(DWORD *)pDataArgs;
			DdeUnaccessData(hArgs);
			GetDocument()->Display(ExplainError());
			DdeFreeDataHandle(hArgs);
			GetDocument()->Display(ExplainError());
			va_end(VarList);
			return;
		}
		else if(strcmp(pFormat, "BL") == 0)	{
			TwoByteBool *pBool;
			pBool = va_arg(VarList, TwoByteBool *);
			*pBool = *(TwoByteBool *)pDataArgs;
			DdeUnaccessData(hArgs);
			GetDocument()->Display(ExplainError());
			DdeFreeDataHandle(hArgs);
			GetDocument()->Display(ExplainError());
			va_end(VarList);
			return;
		}
	}

	//	We are assuming NULL terminated data, since there is more than one
	//		parameter expected.
	HSZ hszArgs = DdeCreateStringHandle(m_dwidInst, pDataArgs,
		CP_WINANSI);
	GetDocument()->Display(ExplainError());
	DdeUnaccessData(hArgs);	
	GetDocument()->Display(ExplainError());
	DdeFreeDataHandle(hArgs);
	GetDocument()->Display(ExplainError());

	int i_ArgNum = 0;
	char *pScan = (char *)pFormat;
	char *pExtract;
	
	//	Loop through the arguments we are going to parse.
	while(pScan && *pScan)	{
		i_ArgNum++;
		pExtract = ExtractArgument(hszArgs, i_ArgNum);
	
		if(0 == strncmp(pScan, "DW", 2))	{
			//	DWORD.
			DWORD *pWord;
			pWord = va_arg(VarList, DWORD *);
			
			//	If there is nothing to scan, use a default value.
			if(pExtract == NULL)	{
				*pWord = 0x0;
			}
			else	{
				sscanf(pExtract, "%lu", pWord);
			}
		}
		else if(0 == strncmp(pScan, "QCS", 3))	{
			//	A quoted CString
			CString *pCS = va_arg(VarList, CString *);
			
			if(pExtract == NULL)	{
				pCS->Empty();
			}
			else	{
				//	Fun thing about a qouted string, is that we need
				//		to compress and '\"' into '"'.
				//	Extractions took off the leading and ending quotes.
				char *pCopy = pExtract;
				while(*pCopy)	{
					if(*pCopy == '\\' && *(pCopy + 1) == '\"')	{
						pCopy++;
					}
					
					*pCS += *pCopy;
					pCopy++;
				}
			}
		}
		else if(0 == strncmp(pScan, "CS", 2))	{
			//	A CString
			CString *pCS = va_arg(VarList, CString *);

			if(pExtract == NULL)	{
				pCS->Empty();
			}
			else	{
				*pCS = pExtract;
			}
		}
		else if(0 == strncmp(pScan, "BL", 2))	{
			//	A boolean
			TwoByteBool *pBool = va_arg(VarList, TwoByteBool *);
			
			if(pExtract == NULL)	{
				*pBool = FALSE;
			}
			else	{
				//	Compare for a TRUE or a FALSE
				if(0 == stricmp(pExtract, "TRUE"))	{
					*pBool = TRUE;
				}
				else	{
					*pBool = FALSE;
				}
			}
		}
		
		//	Go on to the next argument in our format string.
		pScan = SkipToNextArgument(pScan);
		
		//	Free the memory that was used during extraction.
		if(pExtract != NULL)	{
			delete pExtract;
		}
	}
	
	//	Done with variable number of arguments
	va_end(VarList);
	
	//	Free off our created HSZ string.
	DdeFreeStringHandle(m_dwidInst, hszArgs);
	GetDocument()->Display(ExplainError());
}

CString ExplainPoke(HDDEDATA hStatus)	{
	switch((UINT)hStatus)	{
	case DDE_FACK:
		return(CString("ACK "));
	case DDE_FBUSY:
		return(CString("BUSY "));
	case DDE_FNOTPROCESSED:
		return(CString("NOTPROCESSED "));
	}
	
	return(CString("UNKNOWN "));
}

CString ExplainBool(TwoByteBool bFlag)	{
	if(bFlag == FALSE)	{
		return(CString("FALSE "));
	}
	
	return(CString("TRUE "));
}
CString ExplainError()	{
	UINT uError = DdeGetLastError(CDDEObject::m_dwidInst);
	
	switch(uError)	{
	case DMLERR_NO_ERROR:
		return(CString(""));
	case DMLERR_ADVACKTIMEOUT:
		return(CString("ADVACKTIMEOUT"));
	case DMLERR_BUSY:
		return(CString("BUSY"));
	case DMLERR_DATAACKTIMEOUT:
		return(CString("DATAACKTIMEOUT"));		
	case DMLERR_DLL_NOT_INITIALIZED:
		return(CString("DLL_NOT_INITIALIZED"));
	case DMLERR_DLL_USAGE:
		return(CString("DLL_USAGE"));		
	case DMLERR_EXECACKTIMEOUT:
		return(CString("EXECACKTIMEOUT"));
	case DMLERR_INVALIDPARAMETER:
		return(CString("INVALIDPARAMETER"));
	case DMLERR_LOW_MEMORY:
		return(CString("LOW_MEMORY"));
	case DMLERR_MEMORY_ERROR:
		return(CString("DMLERR_MEMORY_ERROR"));
	case DMLERR_NOTPROCESSED:
		return(CString("DMLERR_NOTPROCESSED"));
	case DMLERR_NO_CONV_ESTABLISHED:
		return(CString("NO_CONV_ESTABLISHED"));
	case DMLERR_POKEACKTIMEOUT:
		return(CString("POKEACKTIMEOUT"));
	case DMLERR_POSTMSG_FAILED:
		return(CString("POSTMSG_FAILED"));
	case DMLERR_REENTRANCY:
		return(CString("REENTRANCY"));
	case DMLERR_SERVER_DIED:
		return(CString("SERVER_DIED"));
	case DMLERR_SYS_ERROR:
		return(CString("SYS_ERROR"));
	case DMLERR_UNADVACKTIMEOUT:
		return(CString("UNADVACKTIMEOUT"));
	case DMLERR_UNFOUND_QUEUE_ID:
		return(CString("UNFOUND_QUEUE_ID"));
	}
	
	return(CString("Uknown Error"));
}

//	Client browser interface.
DWORD CDDEObject::WWW_Activate(DWORD dwWindowID, DWORD dwFlags)	{
	DWORD dwActivatedID = 0ul;
	
	//	Make the connection.
	CDDEConversation *pConv = ClientConnect(m_Activate);
	if(pConv != NULL)	{	
		//	Construct the arguments.
		HSZ hszItem = ClientArguments("DW,DW", &dwWindowID, &dwFlags);
		
		//	Do the transaction.
		HDDEDATA hActivatedID = pConv->ClientRequest(hszItem);
		
		//	Convert the return value.
		if(hActivatedID != NULL)	{
			ServerReturned(hActivatedID, "DW", &dwActivatedID);
		}
	
		//	Cut the cord.
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_Activate(%lu, %lu);", dwActivatedID, dwWindowID, dwFlags);
	GetDocument()->Display(aBuf);
	
	return(dwActivatedID);
}
void CDDEObject::WWW_CancelProgress(DWORD dwTransactionID)	{
	//	Make the connection.
	CDDEConversation *pConv = ClientConnect(m_CancelProgress);
	HDDEDATA hStatus = DDE_FNOTPROCESSED;
	
	if(pConv != NULL)	{	
		//	Construct the arguments.
		HSZ hszItem = ClientArguments("DW", &dwTransactionID);
		
		//	Do the transaction.
		hStatus = pConv->ClientPoke(hszItem);
		
		//	Cut the cord.
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "WWW_CancelProgress(%lu);", dwTransactionID);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_Exit()	{
	//	Connect.
	CDDEConversation *pConv = ClientConnect(m_Exit);
	HDDEDATA hStatus = DDE_FNOTPROCESSED;
	
	if(pConv != NULL)	{
		//	Do the transaction.
		CString csTemp = "WWW_Exit";
		HSZ hszItem = ClientArguments("CS", &csTemp);
		hStatus = pConv->ClientPoke(hszItem);
		
		//	Cut the cord.
		pConv->DoDisconnect();
	}
	
	GetDocument()->Display("WWW_Exit()");
}
void CDDEObject::WWW_ListWindows()	{
	CDDEConversation *pConv = ClientConnect(m_ListWindows);
	HDDEDATA hWindows = NULL;
	
	if(pConv != NULL)	{
		CString csTemp = "WWW_ListWindows";
		HSZ hszItem = ClientArguments("CS", &csTemp);
		hWindows = pConv->ClientRequest(hszItem);
		
		pConv->DoDisconnect();
	}
	
	//	Nonstandard data (from the rest of the spec) we are parsing here.
	CString csWindows;
	
	if(hWindows != NULL)	{
		DWORD *pWindows = (DWORD *)DdeAccessData(hWindows, NULL);
		GetDocument()->Display(ExplainError());
		
		//	Go till NULL
		char aBuf[512];
		while(*pWindows)	{
			if(csWindows.IsEmpty() == FALSE)	{
				csWindows += ',';
			}
			sprintf(aBuf, "%lu", *pWindows);
			csWindows += aBuf;
			pWindows++;
		}
		DdeUnaccessData(hWindows);
		GetDocument()->Display(ExplainError());
		DdeFreeDataHandle(hWindows);
		GetDocument()->Display(ExplainError());
		
		if(csWindows.IsEmpty() == FALSE)	{
			csWindows += " ";
		}
	}
	
	GetDocument()->Display(csWindows + "WWW_ListWindows();");
}
void CDDEObject::WWW_Version()	{
	CDDEConversation *pConv = ClientConnect(m_Version);
	DWORD dwVersion = 0;
	
	if(pConv != NULL)	{
		CString csTemp = "WWW_Version";
		HSZ hszItem = ClientArguments("CS", &csTemp);
		HDDEDATA hVersion = pConv->ClientRequest(hszItem);
		
		ServerReturned(hVersion, "DW", &dwVersion);

		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_Version();", dwVersion);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_GetWindowInfo(DWORD dwWindowID)	{
	CDDEConversation *pConv = ClientConnect(m_GetWindowInfo);
	CString csURL;
	CString csTitle;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("DW", &dwWindowID);
		HDDEDATA hWinInfo = pConv->ClientRequest(hszItem);
		
		if(hWinInfo != NULL)	{
			ServerReturned(hWinInfo, "QCS,QCS", &csURL, &csTitle);
		}

		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s,%s WWW_GetWindowInfo(%lu);", (const char *)csURL,
		(const char *)csTitle, dwWindowID);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_OpenURL(CString csURL, CString csSaveAs, DWORD dwWindowID,
	DWORD dwFlags, CString csPostFormData, CString csPostMIMEType,
	CString csProgressServer)	{
	CDDEConversation *pConv = ClientConnect(m_OpenURL);
	DWORD dwServicingID = 0;
	
	if(pConv != NULL)	{
		// Careful to pass optional parameters as NULL here.
		HSZ hszItem = ClientArguments("QCS,QCS,DW,DW,QCS,QCS,CS",
			&csURL,
			csSaveAs.IsEmpty() ? NULL : &csSaveAs,
			&dwWindowID,
			&dwFlags,
			csPostFormData.IsEmpty() ? NULL : &csPostFormData,
			csPostMIMEType.IsEmpty() ? NULL : &csPostMIMEType,
			csProgressServer.IsEmpty() ? NULL : &csProgressServer);
		HDDEDATA hServiceID = pConv->ClientRequest(hszItem);
		
		if(hServiceID != NULL)	{
			ServerReturned(hServiceID, "DW", &dwServicingID);
		}

		pConv->DoDisconnect();
	}
	
	char aBuf[2048];
	sprintf(aBuf, "%lu WWW_OpenURL(%s,%s,%lu,%lu,%s,%s,%s);",
		dwServicingID, (const char *)csURL, (const char *)csSaveAs,
		dwWindowID, dwFlags, (const char *)csPostFormData,
		(const char *)csPostMIMEType, (const char *)csProgressServer);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_ParseAnchor(CString csAbsolute, CString csRelative)	{
	CDDEConversation *pConv = ClientConnect(m_ParseAnchor);
	CString csCombinedURL;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS", &csAbsolute, &csRelative);
		HDDEDATA hCombined = pConv->ClientRequest(hszItem);
		
		if(hCombined != NULL)	{
			ServerReturned(hCombined, "QCS", &csCombinedURL);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_ParseAnchor(%s,%s);", (const char *)csCombinedURL,
		(const char *)csAbsolute, (const char *)csRelative);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_QueryURLFile(CString csFileName)	{
	CDDEConversation *pConv = ClientConnect(m_QueryURLFile);
	CString csURL;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS", &csFileName);
		HDDEDATA hURL = pConv->ClientRequest(hszItem);
		
		if(hURL != NULL)	{
			ServerReturned(hURL, "QCS", &csURL);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[2048];
	sprintf(aBuf, "%s WWW_QueryURLFile(%s);", (const char *)csURL, (const char *)csFileName);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_RegisterProtocol(CString csServer, CString csProtocol)	{
	CDDEConversation *pConv = ClientConnect(m_RegisterProtocol);
	TwoByteBool bRegistered = FALSE;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS", &csServer, &csProtocol);
		HDDEDATA hRegistered = pConv->ClientRequest(hszItem);
		
		if(hRegistered != NULL)	{
			ServerReturned(hRegistered, "BL", &bRegistered);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_RegisterProtocol(%s,%s);", (const char *)ExplainBool(bRegistered),
		(const char *)csServer, (const char *)csProtocol);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_RegisterURLEcho(CString csServer)	{
	CDDEConversation *pConv = ClientConnect(m_RegisterURLEcho);
	HDDEDATA hStatus = DDE_FNOTPROCESSED;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS", &csServer);
		hStatus = pConv->ClientPoke(hszItem);
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "WWW_RegisterURLEcho(%s);", (const char *)csServer);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_RegisterViewer(CString csServer, CString csMIMEType,
	DWORD dwFlags)	{
	CDDEConversation *pConv = ClientConnect(m_RegisterViewer);
	TwoByteBool bRegistered = FALSE;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS,DW", &csServer, &csMIMEType, &dwFlags);
		HDDEDATA hRegistered = pConv->ClientRequest(hszItem);
		
		if(hRegistered != NULL)	{
			ServerReturned(hRegistered, "BL", &bRegistered);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_RegisterViewer(%s,%s,%lu);", (const char *)ExplainBool(bRegistered),
		(const char *)csServer, (const char *)csMIMEType, dwFlags);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_RegisterWindowChange(CString csServer, DWORD dwWindowID)	{
	CDDEConversation *pConv = ClientConnect(m_RegisterWindowChange);
	DWORD dwMonitorID = 0;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,DW", &csServer, &dwWindowID);
		HDDEDATA hMonitoring = pConv->ClientRequest(hszItem);
		
		if(hMonitoring != NULL)	{
			ServerReturned(hMonitoring, "DW", &dwMonitorID);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_RegisterWindowChange(%s,%lu);", dwMonitorID,
		(const char *)csServer, dwWindowID);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_ShowFile(CString csFileName, CString csMimeType,
	DWORD dwWindowID, CString csURL)	{
	CDDEConversation *pConv = ClientConnect(m_ShowFile);
	DWORD dwServicing = 0;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS,DW,QCS", &csFileName, &csMimeType,
			&dwWindowID, &csURL);
		HDDEDATA hServicing = pConv->ClientRequest(hszItem);
		
		if(hServicing != NULL)	{
			ServerReturned(hServicing, "DW", &dwServicing);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_ShowFile(%s,%s,%lu,%s);", dwServicing,
		(const char *)csFileName, (const char *)csMimeType, dwWindowID,
		(const char *)csURL);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_UnRegisterProtocol(CString csServer, CString csProtocol)	{
	CDDEConversation *pConv = ClientConnect(m_UnRegisterProtocol);
	TwoByteBool bUnRegistered = FALSE;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS", &csServer, &csProtocol);
		HDDEDATA hUnRegistered = pConv->ClientRequest(hszItem);
		
		if(hUnRegistered != NULL)	{
			ServerReturned(hUnRegistered, "BL", &bUnRegistered);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_UnRegisterProtocol(%s,%s);", (const char *)ExplainBool(bUnRegistered),
		(const char *)csServer, (const char *)csProtocol);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_UnRegisterURLEcho(CString csServer)	{
	CDDEConversation *pConv = ClientConnect(m_UnRegisterURLEcho);
	HDDEDATA hStatus = DDE_FNOTPROCESSED;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS", &csServer);
		hStatus = pConv->ClientPoke(hszItem);
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "WWW_UnRegisterURLEcho(%s);", (const char *)csServer);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_UnRegisterViewer(CString csServer, CString csMIMEType)	{
	CDDEConversation *pConv = ClientConnect(m_UnRegisterViewer);
	TwoByteBool bUnRegistered = FALSE;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,QCS", &csServer, &csMIMEType);
		HDDEDATA hUnRegistered = pConv->ClientRequest(hszItem);
		
		if(hUnRegistered != NULL)	{
			ServerReturned(hUnRegistered, "BL", &bUnRegistered);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_UnRegisterViewer(%s,%s);", (const char *)ExplainBool(bUnRegistered),
		(const char *)csServer, (const char *)csMIMEType);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_UnRegisterWindowChange(CString csServer, DWORD dwWindowID)	{
	CDDEConversation *pConv = ClientConnect(m_UnRegisterWindowChange);
	TwoByteBool bUnRegistered = FALSE;
	
	if(pConv != NULL)	{
		HSZ hszItem = ClientArguments("QCS,DW", &csServer, &dwWindowID);
		HDDEDATA hUnRegistered = pConv->ClientRequest(hszItem);
		
		if(hUnRegistered != NULL)	{
			ServerReturned(hUnRegistered, "BL", &bUnRegistered);
		}
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_UnRegisterWindowChage(%s,%lu);",
		(const char *)ExplainBool(bUnRegistered),
		(const char *)csServer, dwWindowID);
	GetDocument()->Display(aBuf);
}
void CDDEObject::WWW_WindowChange(DWORD dwWindowID, DWORD dwWindowFlags,
	DWORD dwX, DWORD dwY, DWORD dwCX, DWORD dwCY)	{
	CDDEConversation *pConv = ClientConnect(m_WindowChange);
	HDDEDATA hStatus = DDE_FNOTPROCESSED;
	
	if(pConv != NULL)	{
		//	Be careful here to check optional arguments.
		HSZ hszItem = ClientArguments("DW,DW,DW,DW,DW,DW",
			&dwWindowID,
			&dwWindowFlags,
			(dwWindowFlags & 1) ? &dwX : NULL,
			(dwWindowFlags & 1) ? &dwY : NULL,
			(dwWindowFlags & 1) ? &dwCX : NULL,
			(dwWindowFlags & 1) ? &dwCY : NULL);
		hStatus = pConv->ClientPoke(hszItem);
		
		pConv->DoDisconnect();
	}
	
	char aBuf[512];
	sprintf(aBuf, "WWW_WindowChange(%lu,%lu,%lu,%lu,%lu,%lu);",
		dwWindowID, dwWindowFlags, dwX, dwY, dwCX, dwCY);
	GetDocument()->Display(aBuf);
}

//	Persistance
void CDDEObject::Serialize(CArchive& ar)	{
	if(ar.IsStoring())	{
		//	Add storing code.
	}
	else	{
		//	Add loading code.
	}
	
	//	Serialize the base.
	CObject::Serialize(ar);
}

#ifdef _DEBUG

void CDDEObject::AssertValid() const	{
	//	Call base to AssertValid.
	CObject::AssertValid();
	
	//	Check our instance specific members.
	ASSERT(m_iInstanceCounter);	//	Must always be above 0!
	ASSERT(m_pDoc);	//	Must always have a doc assigned!
	ASSERT(!m_csServiceName.IsEmpty());	//	Must always have a DDE server name!
	ASSERT(m_rIndex);	//	Position in list
}

void CDDEObject::Dump(CDumpContext& dc)	const {
	//	Have the base dump.
	CObject::Dump(dc);
	
	//	Now we dump.
	dc << "Service name: " << m_csServiceName << "\n";
}

#endif // _DEBUG

//	DDEML Callback function
#ifdef _WIN32
HDDEDATA CALLBACK NstestDdeCallBack(UINT type, UINT fmt,
	HCONV hconv, HSZ hsz1, HSZ hsz2, HDDEDATA hData, DWORD dwData1,
	DWORD dwData2)	{
#else
HDDEDATA CALLBACK _export NstestDdeCallBack(UINT type, UINT fmt,
	HCONV hconv, HSZ hsz1, HSZ hsz2, HDDEDATA hData, DWORD dwData1,
	DWORD dwData2)	{
#endif // _WIN32
	
	switch(type)	{				//	ReturnType			Receiver	
	case XTYP_CONNECT:	{			//	XCLASS_BOOL			server
		//	Find the object with the registered service name.
		CDDEObject *pObj = CDDEObject::ResolveService(hsz2);
		if(pObj != NULL)	{
			//	Leave it up to the document to decide wether or not it wants the
			//		conversation.
			return(pObj->ServerConnect(hsz1));
		}
		return(DDEFALSE);
	}
	
	case XTYP_CONNECT_CONFIRM:	{	//	XCLASS_NOTIFICATION	server
		//	Find the object with the registered service name.
		CDDEObject *pObj = CDDEObject::ResolveService(hsz2);
		if(pObj != NULL)	{
			//	The object has already accepted a conversation for the given
			//		topic, so no need to check that.  Let it know it has a real
			//		conversation now, and give it the handle.
			pObj->ServerConnectConfirm(hconv);
		}
		return(NOTHING);
	}
	
	case XTYP_DISCONNECT:	{		//	XCLASS_NOTIFICATION	client/server
		//	Find the object handling the conversation.
		CDDEConversation *pConv = CDDEConversation::ResolveConversation(hconv);
		if(pConv != NULL)	{
			//	We also need to remove the conversation from the list of
			//		conversations managed by some DDE service object.
			CDDEObject::DDEDisconnect(pConv);
		
			//	It's over.
			delete pConv;
		}
		return(NOTHING);
	}
	
	case XTYP_POKE:	{				//	XCLASS_FLAGS		server
		//	Find the object handling the conversation.
		CDDEConversation *pConv = CDDEConversation::ResolveConversation(hconv);
		if(pConv != NULL)	{
			//	Have the object handle the details.
			return(pConv->ServerPoke(hsz1, hsz2, hData));
		}
		return(DDE_FNOTPROCESSED);
	}
	
	case XTYP_REQUEST:	{			//	XCLASS_DATA			server
		//	Find the object handling the conversation.
		CDDEConversation *pConv = CDDEConversation::ResolveConversation(hconv);
		if(pConv != NULL)	{
			return(pConv->ServerRequest(hsz1, hsz2));
		}
		return(NULL);
	}
	}	//	end switch
	
	//	Not handled.
	return(NOTHING);
}
